#include "Cv13_main.h"

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	// Program for objects without textures and without lighting (for the light object)
	nolit_program.Init();
	nolit_program.AddVertexShader("Shaders/nolit_vertex.glsl");
	nolit_program.AddFragmentShader("Shaders/nolit_fragment.glsl");
	nolit_program.Link();

	// Program for the central objects
	object_program.Init();
	object_program.AddVertexShader("Shaders/object_vertex.glsl");
	object_program.AddFragmentShader("Shaders/object_fragment.glsl");
	object_program.Link();

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare the lights

	// Create only a single point light, whose position is defined by the user
	PhongLights_ubo.Init();
	PhongLights_ubo.SetGlobalAmbient(glm::vec3(0.2f));
	PhongLights_ubo.PhongLights.clear();
	PhongLights_ubo.PhongLights.push_back(PhongLight::CreatePointLight(
		glm::vec3(0.0f), glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f)));
	PhongLights_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare materials

	WhiteMaterial_ubo.Init();
	WhiteMaterial_ubo.SetMaterial(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f, 1.0f, 1.0f), true, 200.0f, 1.0f));
	WhiteMaterial_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare all cameras
	
	// Create the basic camera through which we look into the scene
	the_camera.SetCenter(glm::vec3(0.0f));
	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);
	CameraData_ubo.Init();

	//----------------------------------------------
	//--  Prepare geometries
	
	geom_cube = CreateCube();
	geom_sphere = CreateSphere();

	//----------------------------------------------
	//--  Load textures and set their parameters

	// Load Chesterfield set of textures
	chesterfield_color_tex = CreateAndLoadTexture2D(L"../../textures/Chesterfield/ColorMap.jpg");
	glBindTexture(GL_TEXTURE_2D, chesterfield_color_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	chesterfield_normal_tex = CreateAndLoadTexture2D(L"../../textures/Chesterfield/NormalMap.png");
	glBindTexture(GL_TEXTURE_2D, chesterfield_normal_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	chesterfield_height_tex = CreateAndLoadTexture2D(L"../../textures/Chesterfield/HeightMap.png");
	glBindTexture(GL_TEXTURE_2D, chesterfield_height_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	//----------------------------------------------
	//--  Positions of objects in the scene

	// Light's sphere model matrix (its position)
	LightModel_ubo.Init();
	LightModel_ubo.SetMatrix(glm::mat4(1.0f));
	LightModel_ubo.UpdateOpenGLData();

	// Central object's model matrix
	CentralObjectModel_ubo[0].Init();
	CentralObjectModel_ubo[0].SetMatrix(glm::mat4(1.0f));
	CentralObjectModel_ubo[0].UpdateOpenGLData();

	CentralObjectModel_ubo[1].Init();
	CentralObjectModel_ubo[1].SetMatrix(glm::mat4(1.0f));
	CentralObjectModel_ubo[1].UpdateOpenGLData();

	//----------------------------------------------
	//--  Miscelaneous

	// Create the query object
	glGenQueries(1, &RenderTimeQuery);
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	//----------------------------------------------
	//--  Data of the lights

	const float light_distance = 10.0f;
	glm::vec3 light_position = light_distance * glm::vec3(
		cosf(light_pos / 10.0f) * sinf(light_pos),
		sinf(light_pos / 10.0f),
		cosf(light_pos / 10.0f) * cosf(light_pos));

	PhongLights_ubo.PhongLights[0] = PhongLight::CreatePointLight(
		light_position, glm::vec3(0.0f), glm::vec3(1.0f), glm::vec3(1.0f));
	PhongLights_ubo.UpdateOpenGLData();

	LightModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), light_position) * 
		glm::scale(glm::mat4(1.0f), glm::vec3(0.1f)));
	LightModel_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Data of the main camera
	
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 100.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Data of scene objects

	CentralObjectModel_ubo[1].SetMatrix(glm::translate(glm::mat4(1.0f), glm::vec3(object_distance)));
	CentralObjectModel_ubo[1].UpdateOpenGLData();
}

/// Renders the whole frame
void render_scene()
{
	// Start measuring the elapsed time
	glBeginQuery(GL_TIME_ELAPSED, RenderTimeQuery);

	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the framebuffer, clear the color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Set the data of the camera and the lights
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
	PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

	// Render all objects in the scene
	if (object_program.IsValid())
	{
		object_program.Use();
		object_program.Uniform1f("depth_scale_ts", depth_scale);
		object_program.Uniform1i("sample_count", sample_count);

		// Set material and textures
		WhiteMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);
		glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_2D, chesterfield_color_tex);
		glActiveTexture(GL_TEXTURE1);		glBindTexture(GL_TEXTURE_2D, chesterfield_normal_tex);
		glActiveTexture(GL_TEXTURE2);		glBindTexture(GL_TEXTURE_2D, chesterfield_height_tex);

		// Use backface culling for the cubes
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);

		// Render both objects
		geom_cube.BindVAO();
		CentralObjectModel_ubo[0].BindBuffer(DEFAULT_OBJECT_BINDING);
		geom_cube.Draw();
		if (render_second_cube)
		{
			CentralObjectModel_ubo[1].BindBuffer(DEFAULT_OBJECT_BINDING);
			geom_cube.Draw();
		}

		glDisable(GL_CULL_FACE);
	}

	// Render the light's sphere
	if (nolit_program.IsValid())
	{
		nolit_program.Use();

		LightModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		WhiteMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);
		geom_sphere.BindVAO();
		geom_sphere.Draw();
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Stop measuring the elapsed time
	glEndQuery(GL_TIME_ELAPSED);

	// Evaluate the query
	glFinish();					// Wait for OpenGL, don't forget OpenGL is asynchronous
	GLuint64 render_time;
	glGetQueryObjectui64v(RenderTimeQuery, GL_QUERY_RESULT, &render_time);
	render_time_ms = float(render_time) * 1e-6f;
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	light_pos = 0.0f;
	object_distance = 0.5f;
	depth_scale = 0.1f;
	sample_count = 20;
	render_second_cube = false;

	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwDefine("Parameters size='270 150'");
	TwDefine("Parameters valueswidth=80");
	TwDefine("Parameters refresh=0.1");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	// Render time (ms)
	TwAddVarRO(the_gui, "Render time (ms)", TW_TYPE_FLOAT, &render_time_ms, nullptr);

	TwAddVarRW(the_gui, "Light direction", TW_TYPE_FLOAT, &light_pos, "min=-15.0 max=15.0 step=0.03");
	TwAddVarRW(the_gui, "Object distance", TW_TYPE_FLOAT, &object_distance, "min=0.0 max=2.0 step=0.01");
	TwAddVarRW(the_gui, "Depth scale", TW_TYPE_FLOAT, &depth_scale, "min=0.0 max=0.5 step=0.001");
	TwAddVarRW(the_gui, "Sample count", TW_TYPE_INT32, &sample_count, "min=4.0 max=100.0 step=1");
	TwAddVarRW(the_gui, "Render second cube", TW_TYPE_BOOLCPP, &render_second_cube, nullptr);
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Cv13: Parallax Occlusion Mapping");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}